@@ -12,8 +12,9 @@ use namespace::autoclean;
 
 use HPCD::uni::Stage;
 
-with 'HPCI::Group' => { theDriver => 'HPCD::uni' },
-    'HPCD::uni::JobGroup';
+with 'HPCI::Group'    => { theDriver => 'HPCD::uni' },
+	 'HPCI::JobGroup' => { theDriver => 'HPCD::uni' },
+	 HPCI::get_extra_roles('uni', 'group');
 
 =head1 NAME
 
