@@ -688,7 +688,7 @@ other stage or the attempt is aborted).
 			$self->_running_cnt( $self->_running_cnt + 1 );
 			$self->_submitted->{$name} = 1;
 			my ( $now_secs, $now_microsecs ) = gettimeofday;
-			if ($now_secs < $self->_sub_secs + 2) {
+			if ($stage->_has_command && $now_secs < $self->_sub_secs + 2) {
 				my $delta_microsecs =
 					( $now_secs - $self->_sub_secs ) * 1_000_000 +
 					( $now_microsecs - $self->_sub_microsecs );
