@@ -206,10 +206,15 @@ role {
 
 	sub _base_register_status {
 		my $self   = shift;
-		my $status = shift;
+		my $status = shift // 0;
 		my $stath  = $self->stats;
 		$stath->{exit_status} = $status;
-		if ($status & 127) {
+		if ($status =~ /\D/) {
+			$self->killsignal( "internal code failed: $status" );
+			$self->status( $stath->{exit_status} = 126 );
+			;
+		}
+		elsif ($status & 127) {
 			$self->kill( 1 );
 			$self->killsignal( $sig_name[$status & 127] // "Unknown signal #".($status&127) );
 		}
